/**
  *  \file game/playerset.cc
  *  \brief Player-Indexed Data Types
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/playerset.h"
#include "cpluslib/misc.h"
#include "util/nls.h"
#include "game/struct.h"
#include "cpluslib/format.h"

/** Format player set to nice human-readable string. */
string_t
formatPlayerSet(GPlayerSet set)
{
    // hullfunc.pas::FormatRaceMask
    const GPlayerSet allPlayers = GPlayerSet::allUpTo(NUM_PLAYERS) - 0;
    set &= allPlayers;
    if (set == allPlayers) {
        // all players -- nothing to mention
        return string_t();
    }
    if (set.empty()) {
        // no player
        return _("nobody");
    }
    if (set.isUnitSet()) {
        // one player
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            if (set == GPlayerSet(i))
                return format(_("player %d"), i);
    }
    if ((allPlayers - set).isUnitSet()) {
        // all but one player
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            if (set == allPlayers - i)
                return format(_("all but player %d"), i);
    }

    // generic version
    string_t result;
    const char* prefix = _("players ");
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (set.contains(i)) {
            result += prefix;
            result += format("%d", i);
            prefix = ", ";
        }
    }
    return result;
}

string_t
formatPlayerHostSet(GPlayerSet set)
{
    const bool hasHost = set.contains(12);
    const GPlayerSet allPlayers = GPlayerSet::allUpTo(NUM_PLAYERS) - 0;
    set &= allPlayers;

    if (set == allPlayers) {
        // everyone
        if (hasHost) {
            return _("host, all players");
        } else {
            return _("all players");
        }
    }
    if (set.empty()) {
        // nobody
        if (hasHost) {
            return _("host");
        } else {
            return _("nobody");
        }
    }

    string_t formattedPlayers = formatPlayerSet(set);
    if (hasHost) {
        // formattedPlayers can be "all but player X", so put host in front to make it unambiguous
        formattedPlayers = format("host, %s", formattedPlayers);
    }
    return formattedPlayers;
}


/*! \class GPlayerBitMatrix
    \brief Bit Matrix of Players

    This class provides a matrix of bits, such as is used for alliance
    relations.

    For ease of reference, parameters are called "subject" and
    "object" here. For example, "allies.get(2,9)" asks whether player
    2 has offered an alliance to player 9.

    GPlayerBitMatrix permits indexes from 1 to 12.
*/

/** Constructor. */
GPlayerBitMatrix::GPlayerBitMatrix()
{ }

/** Get one bit.
    \param subj Subject
    \param obj  Object
    \return bit value; false if parameters are out of range. */
bool
GPlayerBitMatrix::get(int subj, int obj) const
{
    if (subj <= 0 || subj > int(countof(data)))
        return false;
    if (obj <= 0 || obj > int(countof(data)))
        return false;

    return data[subj-1].contains(obj);
}

/** Set one bit.
    \param subj  Subject
    \param obj   Object
    \param value New value */
void
GPlayerBitMatrix::set(int subj, int obj, bool value)
{
    if (subj <= 0 || subj > int(countof(data)))
        return;
    if (obj <= 0 || obj > int(countof(data)))
        return;

    if (value)
        data[subj-1] += obj;
    else
        data[subj-1] -= obj;
}

/** Get one row.
    \param subj Subject
    \return set of all objects for this subject. */
GPlayerSet
GPlayerBitMatrix::getRow(int subj) const
{
    if (subj <= 0 || subj > int(countof(data)))
        return GPlayerSet();
    return data[subj-1];
}

/** Clear this set. */
void
GPlayerBitMatrix::clear()
{
    for (int i = 0; i < int(countof(data)); ++i)
        data[i].clear();
}
