/**
  *  \file game/univ.h
  *  \brief Universe: A Container for (Almost) Everything
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UNIV_H
#define PCC_V2_GAME_UNIV_H

#include "cpluslib/sigslot.h"
#include "cpluslib/string.h"
#include "game/coord.h"
#include "game/drawing.h"
#include "game/expl.h"
#include "game/struct.h"
#include "game/timestamp.h"
#include "game/ufo.h"
#include "game/mine.h"
#include "game/playerset.h"
#include "game/shiptype.h"
#include "game/planettype.h"
#include "game/fleet.h"
#include "game/ion.h"
#include "cpluslib/ptrvector.h"

class GPlanet;
class GShip;
class GIonStorm;
class GMessageInformation;

//! Flags for GUniverse::getLocationName().
enum {
    locs_Orbit   = 1,           ///< Show "Orbit of" for planet names.
    locs_Verbose = 2,           ///< Be more verbose.
    locs_Ships   = 4,           ///< Show a ship name if applicable.
    locs_WW      = 8,           ///< Show planet name if point is in warp well.
    locs_NoSpace = 16           ///< Show nothing at all when in deep space.
};

class GUniverse {
 public:
    GUniverse();
    ~GUniverse();

    // Initialisation
    void postprocess(GPlayerSet playing_set, GPlayerSet data_set, GObject::Playability playability);
    void recomputeOrbitFlags();

    // Turn number
    void setTurnNumber(int nr);
    int  getTurnNumber() const;
    void setTimestamp(GTimestamp ts);
    const GTimestamp& getTimestamp() const;

    bool isData(int player_nr) const
        { return data_set.contains(player_nr); }

    // Object accessors
    GPlanet&          getPlanet(int id);
    const GPlanet&    getPlanet(int id) const;
    bool              isValidPlanetId(int id) const;
    GShip&            getShip(int id);
    const GShip&      getShip(int id) const;
    bool              isValidShipId(int id) const;
    GIonStorm&        getIonStorm(int sid);
    const GIonStorm&  getIonStorm(int sid) const;
    bool              isValidIonStormId(int id) const;

    GDrawingContainer& getDrawings()
        { return drawings; }
    GDrawingContainer const& getDrawings() const
        { return drawings; }

    GExplosionContainer& getExplosions()
        { return explosions; }
    GExplosionContainer const& getExplosions() const
        { return explosions; }

    // Location accessors
    int       getPlanetAt(GPoint pt) const;
    int       getPlanetAt(GPoint pt, bool gravity) const;
    int       getGravityPlanetAt(GPoint pt) const;
    int       getAnyShipAt(GPoint pt) const;
    int       getNumberOfPlayedShipsAt(GPoint pt) const;
    string_t  getLocationName(GPoint pt, int flags) const;
    bool      isShipTowed(int sid) const;

    // Updates
    void doScreenUpdates();
    void markChanged();

    void addMessageInformation(const GMessageInformation& msgi, int msgNr);

    Signal0 sig_pre_update;
    Signal0 sig_universe_changed;

    GPlayedShipType    ty_played_ships;
    GAnyShipType       ty_any_ships;
    GHistoryShipType   ty_history_ships;
    GPlayedPlanetType  ty_played_planets;
    GAnyPlanetType     ty_any_planets;
    GPlayedBaseType    ty_played_bases;
    GFleetType         ty_fleets;
    GUfoType           ty_ufos;
    GIonStormType      ty_ion_storms;
    GMinefieldType     ty_minefields;

 private:
    void init();

    GPlayerSet   data_set;
    ptr_vector<GPlanet>   planets;
    ptr_vector<GShip>     ships;
    ptr_vector<GIonStorm> storms;
    GDrawingContainer     drawings;
    GExplosionContainer   explosions;
    bool         universe_changed;

    int turn_nr;
    GTimestamp timestamp;
};

inline void
GUniverse::setTurnNumber(int nr)
{
    turn_nr = nr;
}

inline int
GUniverse::getTurnNumber() const
{
    return turn_nr;
}

inline void
GUniverse::setTimestamp(GTimestamp ts)
{
    timestamp = ts;
}

inline const GTimestamp&
GUniverse::getTimestamp() const
{
    return timestamp;
}

inline void
GUniverse::markChanged()
{
    universe_changed = true;
}

#endif
