/**
  *  \file ui/widgets/tooltip.cc
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/tooltip.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/window.h"
#include "ui/uiglobal.h"

/******************************* UITooltip *******************************/

UITooltip::UITooltip(string_t content, GfxPoint anchor)
    : content(content),
      anchor(anchor),
      top(false)
{
    computePosition();
}

UITooltip::~UITooltip()
{ }

void
UITooltip::computePosition()
{
    // Figure out size of text
    int wrap   = ui_root->getExtent().w * 9 / 10;
    int lines  = countLines(*fonts[FONT_SMALL], wrap, content);
    int height = lines * font_heights[FONT_SMALL];
    int width  = getTextWidthFormatted(*fonts[FONT_SMALL], wrap, content);

    int grossHeight = OUT + height + FRAME;

    // Figure out layout
    GfxRect r;
    if (anchor.y < grossHeight) {
        --anchor.x, ++anchor.y;
        r.y = anchor.y;
        top = false;
    } else {
        --anchor.x, --anchor.y;
        r.y = anchor.y - grossHeight;
        top = true;
    }
    r.h = grossHeight;

    r.x = anchor.x + OUT - (width + FRAME);
    r.w = width + FRAME;
    r.moveInto(ui_root->getExtent());

    setExtent(r);
}

bool
UITooltip::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key || event.type == ev_Mouse) {
        stop(0);
        ungetEvent(event);
        return true;
    } else {
        return UIFixedWidget::handleEvent(event, second_pass);
    }
}

void
UITooltip::drawContent(GfxCanvas& can)
{
    // Size
    GfxRect r = getExtent();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_SMALL]);

    if (top) {
        /*
          ,---------------------. ---
          |                     |
          |                     |
          `--------------.  .---' ---
                          \ |     OUT
                           \|     ---
         */
        // Body that will contain text
        drawSolidBar(ctx, GfxRect(r.x, r.y, r.w, r.h-OUT), COLOR_TOOLTIP);

        // Frame
        ctx.setColor(COLOR_TOOLTIP_LIGHT);
        drawHLine(ctx, r.x, r.y,   r.getRightX()-1);
        drawVLine(ctx, r.x, r.y+1, r.getBottomY() - OUT);

        ctx.setColor(COLOR_TOOLTIP_SHADE);
        drawVLine(ctx, r.getRightX()-1, r.y,                  r.getBottomY() - OUT);
        drawHLine(ctx, r.x,             r.getBottomY() - OUT, anchor.x - OUT);
        drawHLine(ctx, anchor.x,        r.getBottomY() - OUT, r.getRightX()-1);
        drawVLine(ctx, anchor.x,        r.getBottomY() - OUT, r.getBottomY()-1);

        // Pointer
        ctx.setColor(COLOR_TOOLTIP);
        for (int i = 0; i < OUT; ++i) {
            can.drawPixel(GfxPoint(anchor.x - OUT + i, r.getBottomY() - OUT + i), standard_colors[COLOR_TOOLTIP_SHADE], GFX_ALPHA_OPAQUE);
            if (i < OUT-1) {
                drawHLine(ctx,
                          anchor.x - OUT + i + 1,
                          r.getBottomY() - OUT + i,
                          anchor.x - 1);
            }
        }

        // Text
        ctx.setColor(COLOR_TOOLTIP_TEXT);
        outTextFormatted(ctx, r.x + FRAME/2, r.y + FRAME/2, r.w, content);
    } else {
        /*
                          /|      ---
                         / |      OUT
          ,-------------'  '----. ---
          |                     |
          |                     |
          `---------------------' ---
         */
        // Body that will contain text
        drawSolidBar(ctx, GfxRect(r.x, r.y+OUT, r.w, r.h-OUT), COLOR_TOOLTIP);

        // Frame
        ctx.setColor(COLOR_TOOLTIP_LIGHT);
        drawHLine(ctx, r.x,      r.y+OUT,    anchor.x-OUT);
        drawHLine(ctx, anchor.x, r.y+OUT,    r.getRightX()-1);
        drawVLine(ctx, r.x,      r.y+OUT,    r.getBottomY()-1);

        ctx.setColor(COLOR_TOOLTIP_SHADE);
        drawVLine(ctx, r.getRightX()-1, r.y + OUT+1,      r.getBottomY()-1);
        drawHLine(ctx, r.x,             r.getBottomY()-1, r.getRightX()-1);
        drawVLine(ctx, anchor.x,        r.y,              r.y + OUT);

        // Pointer
        ctx.setColor(COLOR_TOOLTIP);
        for (int i = 0; i < OUT; ++i) {
            can.drawPixel(GfxPoint(anchor.x - i - 1, r.y + i), standard_colors[COLOR_TOOLTIP_LIGHT], GFX_ALPHA_OPAQUE);
            if (i > 0) {
                drawHLine(ctx,
                          anchor.x - i,
                          r.y + i,
                          anchor.x - 1);
            }
        }

        // Text
        ctx.setColor(COLOR_TOOLTIP_TEXT);
        outTextFormatted(ctx, r.x + FRAME/2, r.y + OUT + FRAME/2, r.w, content);
    }

    can.update(r);
}

/**************************** UITooltipControl ***************************/

/** Construct tooltip control. */
UITooltipControl::UITooltipControl(UIBaseWidget& b)
    : b(b),
      mouse_inside(false),
      mouse_pos(0, 0),
      mouse_counter(0)
{ }

/** Check event. Should be called from b's handleEvent(). Checks and processes the
    provided event. Returns true iff the tooltip should be displayed. Otherwise,
    events are only scanned and should be processed normally.

    This will call setSelectedWidget() on the controlled widget if the mouse is inside it. */
bool
UITooltipControl::checkEvent(const UIEvent& event)
{
    bool result = false;
    if (event.type == ev_Tick) {
        /* Tick: count down */
        if (mouse_inside && b.hasState(UIBaseWidget::st_Selected)) {
            if (++mouse_counter > TIMER_HZ/2) {
                result = true;
            }
        } else {
            mouse_inside = false;
        }
    }
    if (event.type == ev_Mouse) {
        /* Mouse: reset count-down, and track position */
        mouse_counter = 0;
        mouse_pos     = GfxPoint(event.mouse.x, event.mouse.y);
        mouse_inside  = b.isInside(mouse_pos);
        if (mouse_inside)
            b.setSelectedWidget();
    }
    if (event.type == ev_Key) {
        /* Keyboard: reset count-down */
        mouse_counter = 0;
    }
    return result;
}

/** Check state. Should be called from b's onStateChange(). */
void
UITooltipControl::checkState()
{
    /* If this function is not called by the user, focus change will not prevent
       tooltips from appearing. For example, when the mouse hovers over a widget,
       keypresses should inhibit tooltips. They do so by obtaining logical focus,
       i.e. the st_Selected bit, causing the previous st_Selected widget to lose
       it. In addition, this catches widgets that steal mouse events when the
       mouse is moved very fast from the controlled widget to another one. */
    if (!b.hasState(UIBaseWidget::st_Selected)) {
        mouse_inside = false;
    }
}

/** Get position where tooltip shall appear. */
GfxPoint
UITooltipControl::getTooltipPosition() const
{
    return mouse_pos;
}

void
doTooltip(string_t content, GfxPoint anchor)
{
    UITooltip tt(content, anchor);
    tt.setState(UIWidget::st_Modal, true);
    ui_root->add(tt);
    ui_root->run();
    ui_root->close(tt);
}
